@@ -7,16 +7,20 @@
 #include <rokubimini_ethercat/TxPdoZ.hpp>
 #include <rokubimini_ethercat/sdo/RxSdoCalibration.hpp>
 
+#include <fstream>
+#include <thread>
 namespace rokubimini
 {
 namespace ethercat
 {
-RokubiminiEthercatSlave::RokubiminiEthercatSlave(const std::string& name, soem_interface::EthercatBusBase* bus,
+RokubiminiEthercatSlave::RokubiminiEthercatSlave(const std::string& name,
+                                                 rokubimini::soem_interface::EthercatBusBase* bus,
                                                  const uint32_t address, const PdoTypeEnum pdoTypeEnum)
-  : soem_interface::EthercatSlaveBase(bus, address)
+  : rokubimini::soem_interface::EthercatSlaveBase(bus, address)
   , name_(name)
   , pdoTypeEnum_(pdoTypeEnum)
   , currentPdoTypeEnum_(PdoTypeEnum::NA)
+  , isRunning_{ true }
 {
   PdoInfo pdo_info;
 
@@ -34,6 +38,9 @@ bool RokubiminiEthercatSlave::startup()
   return configurePdo(pdoTypeEnum_);
 }
 
+void RokubiminiEthercatSlave::preSetupConfiguration()
+{
+}
 void RokubiminiEthercatSlave::updateRead()
 {
   std::lock_guard<std::recursive_mutex> lock(mutex_);
@@ -46,7 +53,7 @@ void RokubiminiEthercatSlave::updateRead()
       bus_->readTxPdo(address_, tx_pdo_a);
       {
         reading_.getWrench().header.stamp = update_stamp;
-        reading_.getWrench().header.frame_id = name_;
+        reading_.getWrench().header.frame_id = name_ + "_wrench";
         reading_.getWrench().wrench.force.x = tx_pdo_a.forceX_;
         reading_.getWrench().wrench.force.y = tx_pdo_a.forceY_;
         reading_.getWrench().wrench.force.z = tx_pdo_a.forceZ_;
@@ -55,7 +62,7 @@ void RokubiminiEthercatSlave::updateRead()
         reading_.getWrench().wrench.torque.z = tx_pdo_a.torqueZ_;
 
         reading_.getImu().header.stamp = update_stamp;
-        reading_.getImu().header.frame_id = name_;
+        reading_.getImu().header.frame_id = name_ + "_imu";
         reading_.getImu().angular_velocity.x = tx_pdo_a.angularRateX_ * DEG_TO_RAD;
         reading_.getImu().angular_velocity.y = tx_pdo_a.angularRateY_ * DEG_TO_RAD;
         reading_.getImu().angular_velocity.z = tx_pdo_a.angularRateZ_ * DEG_TO_RAD;
@@ -71,7 +78,7 @@ void RokubiminiEthercatSlave::updateRead()
         reading_.setForceTorqueSaturated(tx_pdo_a.forceTorqueSaturated_ != 0);
 
         reading_.getTemperature().header.stamp = update_stamp;
-        reading_.getTemperature().header.frame_id = name_;
+        reading_.getTemperature().header.frame_id = name_ + "_temp";
         reading_.getTemperature().temperature = tx_pdo_a.temperature_;
         reading_.getTemperature().variance = 0;  // unknown variance
       }
@@ -253,6 +260,25 @@ bool RokubiminiEthercatSlave::setSensorCalibration(const calibration::SensorCali
 
   return success;
 }
+bool RokubiminiEthercatSlave::setConfigMode()
+{
+  setState(EC_STATE_PRE_OP);
+  // Sleep for some time to give the slave time to execute the pre-op cb
+  std::this_thread::sleep_for(std::chrono::milliseconds(500));
+  if (!waitForState(EC_STATE_PRE_OP))
+  {
+    ROS_ERROR("[%s] Slave failed to switch to PREOP state", name_.c_str());
+    return false;
+  }
+  return true;
+}
+
+bool RokubiminiEthercatSlave::setRunMode()
+{
+  setState(EC_STATE_SAFE_OP);
+  setState(EC_STATE_OPERATIONAL);
+  return true;
+}
 
 bool RokubiminiEthercatSlave::saveConfigParameter()
 {
@@ -335,5 +361,59 @@ bool RokubiminiEthercatSlave::configurePdo(const PdoTypeEnum pdoTypeEnum)
   return true;
 }
 
+bool RokubiminiEthercatSlave::readFileToBuffer(const std::string& filePath)
+{
+  // Read the file into the file buffer
+  std::ifstream file(filePath, std::ios::binary);
+  std::string file_buffer;
+  if (file.is_open())
+  {
+    int ch;
+    while ((ch = file.get()) != std::istream::traits_type::eof())
+    {
+      file_buffer.push_back(static_cast<char>(ch));
+    }
+    fileBuffer_ = const_cast<char*>(file_buffer.c_str());
+    fileSize_ = static_cast<int>(file_buffer.size());
+    file.close();
+    if (fileSize_ > MAX_FILE_SIZE_FIRMWARE)
+    {
+      ROS_ERROR("[%s] File is too big.", name_.c_str());
+      return false;
+    }
+  }
+  else
+  {
+    ROS_ERROR_STREAM("[" << getName() << "] "
+                         << "Failed to open file " << filePath);
+    return false;
+  }
+  ROS_DEBUG_STREAM("[" << getName() << "] "
+                       << "The firmware was read successfully. Size of file buffer: " << fileSize_);
+  return true;
+}
+
+bool RokubiminiEthercatSlave::firmwareUpdate(const std::string& filePath, const std::string& fileName,
+                                             const uint32_t& password)
+{
+  bool success;
+  std::lock_guard<std::recursive_mutex> lock(mutex_);
+  if (!readFileToBuffer(filePath))
+  {
+    ROS_ERROR_STREAM("[" << getName() << "] "
+                         << "Could not read file in path " << filePath << ".");
+    return false;
+  }
+  success = bus_->writeFirmware(address_, fileName, password, fileSize_, fileBuffer_);
+  if (!bus_->isRunning())
+  {
+    isRunning_ = false;
+  }
+  if (!success)
+  {
+    ROS_ERROR("[%s] Flashing was not successful.", name_.c_str());
+  }
+  return success;
+}
 }  // namespace ethercat
-}  // namespace rokubimini
\ No newline at end of file
+}  // namespace rokubimini
